#include "dobpf.h"


static int bpf_fd = -1;

/* Compute the easy part of the checksum on a range of bytes. */
/*
u_int32_t in_cksum (buf, nbytes, sum)
	unsigned char *buf;
	int nbytes;
	u_int32_t sum;
{
	int i;



	for (i = 0; i < (nbytes & ~1); i += 2) {
		sum += (u_int16_t) ntohs(*((u_int16_t *)(buf + i)));

		if (sum > 0xFFFF)
			sum -= 0xFFFF;
	}	


	if (i < nbytes) {

		sum += buf [i] << 8;

		if (sum > 0xFFFF)
			sum -= 0xFFFF;
	}
	printf("The Value Of CheckSum is------------------------------------------------------------------------> %x \n",sum);
	return sum;
}
*/

int in_cksum(addr, len)
	u_short *addr;
	int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	u_short answer = 0;

	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}


	if (nleft == 1) {
		*(u_char *)(&answer) = *(u_char *)w ;
		sum += answer;
	}


	sum = (sum >> 16) + (sum & 0xffff);	
	sum += (sum >> 16);			
	answer = ~sum;				
	return(answer);
}

/*
 * append mac/ip/udp headers and issue packet on current bpf interface
 * note: bpf_fd found internally for reasons of convenience.
 *
 * DHCP broadcast is the current consumer of this function.
 *
 * note: if macdst is NULL, we use limited broadcast (all ffs).
 * note: ipsrc, ipdst assumed in network byte order on input.
 * 	 srcport/dstport assumed to NOT be in network byte order.
 *
 * function currently uses writev, but doesn't need it.
 * Some code ideas borrowed from ISC dhcp client, but in the main
 * based on BSD kernel code.
 * 
 */

writebpf_udp(char *packet, int length, 
		char *macsrc, char *macdst,
		struct in_addr ipsrc, struct in_addr ipdst, 
		int srcport, int dstport)
{
	int bufoffset=0;
	unsigned char buf[1514];
	struct iovec iov[2];
	int rc; 
	extern int debug;
	struct ether_header *eh;
	struct udpiphdr *ui;

	/* copy payload into same packet buffer, skipping layers 2/3/4
	*/
	char *pkt = (char *) &buf[0];
	pkt = pkt + sizeof(struct ether_header) + sizeof(struct udpiphdr);	
	bcopy(packet, pkt, length);	/* copy data part into pkt part */

	/* build the ethernet header first.
	*/
	eh = (struct ether_header *)&buf[0];
	
	memcpy(eh->ether_shost, macsrc, 6);

	if (macdst == NULL) {
	  memset(eh->ether_dhost, 0xff, 6);
	}
	else {
	  memcpy(eh->ether_dhost, macsrc, 6);
	}
	eh->ether_type = htons(ETHERTYPE_IP);
	bufoffset = ETHER_HDRLEN;	


        /*
         * Fill in mbuf with extended UDP header
         * and addresses and length put into network format.
         */
        ui = (struct udpiphdr *) &buf[bufoffset];
        bzero(ui->ui_i.ih_x1, sizeof(ui->ui_i.ih_x1));	/* zero out-non csum 1st 9 ip bytes */
        ui->ui_i.ih_pr = IPPROTO_UDP;
        ui->ui_i.ih_len = htons((u_short)length + sizeof (struct udphdr));
        ui->ui_i.ih_src = ipsrc;	/* already byte-swapped */
        ui->ui_i.ih_dst = ipdst;	/* already byte-swapped */
	
        ui->ui_u.uh_sport = htons(srcport);
        ui->ui_u.uh_dport = htons(dstport);
        ui->ui_u.uh_ulen = ui->ui_i.ih_len;
	
        /*
         * do checksum
         */
        ui->ui_u.uh_sum = 0;
        if ( ( ui->ui_u.uh_sum = in_cksum(ui, sizeof (struct udpiphdr) + length)) == 0)
	  ui->ui_u.uh_sum = 0xffff;
	
	/* udp header is finished, ip header needs work
	 */
	
        ((struct ip *)ui)->ip_v = 4;
        ((struct ip *)ui)->ip_hl = 5;
        ((struct ip *)ui)->ip_tos = IPTOS_LOWDELAY;
        ((struct ip *)ui)->ip_len = htons(sizeof (struct udpiphdr) + length);
        ((struct ip *)ui)->ip_id = 0;
        ((struct ip *)ui)->ip_off = 0;
        ((struct ip *)ui)->ip_ttl = 16;
	/* one more cksum calculation on ip hdr this time
	 */
        ((struct ip *)ui)->ip_sum = 0;
        ((struct ip *)ui)->ip_sum = in_cksum ((struct ip *)ui, sizeof(struct ip));
	
	bufoffset += sizeof(struct udpiphdr);
	bufoffset += length;	/* now we have total length */
	if (debug) {
	  printf("bpf write, total length %d\n", bufoffset);
	}
	
	iov[0].iov_base = buf;
	iov[0].iov_len = bufoffset;
	printf("Sending Broadcast Using bpf_fd = %d \n",bpf_fd);
	rc = writev(bpf_fd, iov, 1);
	if (debug) {
	  printf("bpf write returns %d\n", rc);
	}
	return(rc);
}

/*
 * set the device for bpf output.  This is only used by
 * agents, which do not use the bpf to read packets.
 */
aim_bpf(char *device, char *ebuf)
{
	int rc;
 	struct ifreq ifr;
	int fd;
	struct bpf_program p;

	fd = bpf_fd;

	/* set the network device
	*/
        (void)strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
        if (ioctl(fd, BIOCSETIF, (caddr_t)&ifr) < 0) {
                sprintf(ebuf, "%s: %s", device, strerror(errno));
		close(fd);
		return(-1);
        }
	return(0);
} 



static int map_arphrd_to_dlt(int arptype)
{
        switch (arptype) {
        case ARPHRD_ETHER:
        case ARPHRD_METRICOM:
        case ARPHRD_LOOPBACK:   return DLT_EN10MB;
        case ARPHRD_EETHER:     return DLT_EN3MB;
        case ARPHRD_AX25:       return DLT_AX25;
        case ARPHRD_PRONET:     return DLT_PRONET;
        case ARPHRD_CHAOS:      return DLT_CHAOS; 
#ifndef ARPHRD_IEEE802_TR
#define ARPHRD_IEEE802_TR 800   /* From Linux 2.4 */
#endif
        case ARPHRD_IEEE802_TR:
        case ARPHRD_IEEE802:    return DLT_IEEE802;
        case ARPHRD_ARCNET:     return DLT_ARCNET;
        case ARPHRD_FDDI:       return DLT_FDDI;

#ifndef ARPHRD_ATM  /* FIXME: How to #include this? */
#define ARPHRD_ATM 19 
#endif
        case ARPHRD_ATM:        return DLT_ATM_CLIP;

        case ARPHRD_PPP:
        /* Not sure if this is correct for all tunnels, but it
         * works for CIPE */
        case ARPHRD_TUNNEL:
#ifndef ARPHRD_SIT
#define ARPHRD_SIT 776  /* From Linux 2.2.14 */
#endif
	case ARPHRD_SIT:
        case ARPHRD_CSLIP:
        case ARPHRD_SLIP6:
        case ARPHRD_CSLIP6:
        case ARPHRD_SLIP:       return DLT_RAW;
        }

        return -1;
}
/*
 *  Get the hardware type of the given interface as ARPHRD_xxx constant.
 */
static int
iface_get_arptype(int fd, const char *device, char *ebuf) 
{
        struct ifreq    ifr;

        memset(&ifr, 0, sizeof(ifr));
        strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));

        if (ioctl(fd, SIOCGIFHWADDR, &ifr) == -1) {
                snprintf(ebuf, 1000,
                         "ioctl: %s", strerror(errno));
                return -1;
        }

        return ifr.ifr_hwaddr.sa_family;
}

int bpf_needed(int sock,char *device)
{
  int arptype,linktype;
  char *ebuf = (char *) malloc(1000);
  arptype = iface_get_arptype(sock, device, ebuf);
  linktype =  map_arphrd_to_dlt(arptype);
  
  if (linktype == DLT_EN10MB)
      return 1;

  return 0;
}

bpf_close(int sock)
{
        close(sock);
} 

/*
 *  Bind the socket associated with FD to the given device. 
 */
static int
iface_bind(int fd, int ifindex)
{
	struct sockaddr_ll	sll;

	memset(&sll, 0, sizeof(sll));
	sll.sll_family		= AF_PACKET;
	sll.sll_ifindex		= ifindex;
	sll.sll_protocol	= htons(ETH_P_ALL);

	if (bind(fd, (struct sockaddr *) &sll, sizeof(sll)) == -1) 
	  {
	    printf("iface_get_id %s \n",strerror(errno));
	    return -1;
	  }
	return 0;
}

/*
 *  Return the index of the given device name. Fill ebuf and return 
 *  -1 on failure.
 */
static int
iface_get_id(int fd, const char *device)
{
  struct ifreq	ifr;
  
  memset(&ifr, 0, sizeof(ifr));
  strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
  
  if (ioctl(fd, SIOCGIFINDEX, &ifr) == -1) 
    {
      printf("iface_get_id %s \n",strerror(errno));
      return -1;
    }
  return ifr.ifr_ifindex;
}

int openDes(char *device)
{
  int sock_fd = -1;
  int device_id = -1;

  sock_fd = device ? 
    socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL))
    : socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_ALL));
  
  if (sock_fd == -1) 
    {
      printf("Err Opening Socket %s \n",strerror(errno));
      return -1;
    }

  device_id = iface_get_id(sock_fd, device);
  if (device_id == -1)
    return -1;
  
  if (iface_bind(sock_fd, device_id) == -1)
    return -1;
  
  return sock_fd;
}

int init_bpf(char *device,char *errbuff)
{
  handle_t *handle = (handle_t *) malloc(sizeof(handle_t));
  int sock;
  
  if (handle == NULL) 
    {
      printf("Malloc Err %s \n",strerror(errno));
      return -1;
    }
  memset(handle, 0, sizeof(*handle));
  if (!device || strcmp(device, "any") == 0) 
    {
      device			= NULL;
      handle->device	= strdup("any");
    } 
  else
    handle->device	= strdup(device);
  
  if ( (sock = openDes(device)) < 0)
    return -1;
  
  bpf_fd = sock;
  return sock;
}

/*
 * return directly: pointer to remainder of buffer (ICMP portion)
 * inputs:
 *	buf buffer read from bpf
 *	insize - size of that buffer/packet
 * outputs:
 *	outsize - size LEFT after we skip numerous things
 *	macflag - 1 if valid MAC seen
 *	mac - 6 byte mac address for MAC source
 *	iplen - length of this ip packet
 */

char *
bpf_process_read(char *buf, int insize, int *outsize, int *macflag, char *mac, int *iplen)
{
	char *bp = buf;
	int hdrlen;
	int phdrlen = 0;
	struct ether_header *eh;
	unsigned short type;
	struct llc *l;
	struct ip *ipp;

	*macflag = 0;


#define bhp ((struct bpf_hdr *)bp)

	hdrlen = bhp->bh_hdrlen;
	bp += hdrlen;
	insize -= hdrlen;

	eh = (struct ether_header *) bp;

	type = ntohs(eh->ether_type);
	insize -= ETHER_HDRLEN;
	bp += ETHER_HDRLEN;

	memcpy(mac, eh->ether_shost, 6);
	*macflag = 1;

	ipp = (struct ip *) bp;
	*iplen =  (int) ntohs(ipp->ip_len);
	insize -= *iplen;

	*outsize = insize;	
	
	return(bp);
}

int install_bpf_program(handle_t *p, struct bpf_program *fp)
{
  size_t prog_size;
  
  prog_size = sizeof(*fp->bf_insns) * fp->bf_len;
  p->fcode.bf_len = fp->bf_len;
  p->fcode.bf_insns = (struct bpf_insn *) malloc(prog_size);
  p->fcode.bf_insns = (struct bpf_insn *) malloc(prog_size);
  memcpy(p->fcode.bf_insns, fp->bf_insns, prog_size);
  return 0;
}

int handleBpfInput(int sock)
{
  char               *buffer = (char *) malloc( sizeof(char) * 32768);
  struct sockaddr_ll from;
  struct sll_header  *hdrp;
  socklen_t          fromlen;
  int                packet_len;
  int                i;
  int                caplen;
  int                buffLength =  32768;
  handle_t           *handle = (handle_t *) malloc(sizeof(handle_t));
  struct ether_header *eptr;
  static struct bpf_insn total_insn[]  = {
    BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 12 ),
    BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K,2048, 0, 3 ), 
    BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 23 ),
    BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 1, 0, 1),
    BPF_STMT(BPF_RET+BPF_K, 8192),
    BPF_STMT(BPF_RET+BPF_K,0)
  };
  
  static struct bpf_program f = { sizeof(total_insn) / sizeof(struct 
				bpf_insn), total_insn };


  if (buffer == NULL )
    {
      printf("Invalid Pointer 'buffer':read_packet()\n");
      return -1;	
    }
  
  do {
    fromlen = sizeof(from);
    packet_len = recvfrom(
			  sock, buffer, buffLength, MSG_TRUNC,
			  (struct sockaddr *) &from, &fromlen);
  } while (packet_len == -1 && errno == EINTR); 
  
  if (packet_len == -1) {
    if (errno == EAGAIN)
      {
	printf("No Packet There \n");
	return 0;       
      }
    else {
      printf("Err Reading Packet %s \n",strerror(errno));
      return -1;
    }
  }    

  caplen = packet_len;
  if (caplen > buffLength)
    caplen = buffLength;

  if (install_bpf_program(handle, &f) < 0)
    {
      printf("Unable to Install Filter program \n");
      return -1;
    }
	
  if (bpf_filter(handle->fcode.bf_insns, buffer,
		 packet_len, caplen) == 0)
    {
      //rejected by filter
      return 0;
    }
  else
    {
      char                  mac[6];
      int                   iplen;
      char                  *p;
      int                   outsize;
      char                  *bp;
      const struct my_ip*   ip;
      int                   macflag;
      u_int                 protocol = -1;
      
      p          = (char  *) (buffer + sizeof(struct ether_header));
      iplen      = (int) ntohs(((struct ip *) p )->ip_len);
      macflag    = 1;
      eptr       = (struct ether_header *) buffer;
      memcpy(mac,eptr->ether_shost,6);

      handleIcmpInput(p,iplen,macflag,mac);

      printf("Accepted By the Filter\n");
      printf("-----------------------------------------------------------------------------------\n");
      printf("Ethernet Header Source: %s \n",ether_ntoa(((struct ether_header *)eptr)->ether_shost));
      printf("Ethernet Header Destination: %s\n",ether_ntoa(((struct ether_header *)eptr)->ether_dhost));
      ip         = (struct my_ip *) (buffer + sizeof(struct ether_header));
      printf("IP Source: %s \n",inet_ntoa(((struct ip *)ip)->ip_src));
      printf("IP Destination: %s \n",inet_ntoa(((struct ip *)ip)->ip_dst));
      printf("IP Length Of The IP Packet: %d \n",iplen);
      printf("The Protocol field Value is %d \n",((struct ip *)ip)->ip_p);
      printf("-----------------------------------------------------------------------------------\n");
    }
  return 0;
}

#ifdef TESTMAIN
main()
{
  int sock;
  char *bp;
  char *p; 
  int rc;
  char *expr = "icmp";
  char *errbuf;
  int fd;
  char buf[BPFBUFSIZE];
  int macflag;
  char mac[6];
  int outsize;
  int ii;
  int iplen;
  int cnt = 10;
  int buffLength =  32768;
  char *buffer = (char *) malloc( sizeof(char) * 32768);
  sock = init_bpf("eth0",errbuf);
  
  for(ii=0;ii<cnt;ii++)
    handleBpfInput(sock);
}
#endif


